""" Confiture schema validator.
"""


class ValidationError(Exception):

    """ Exception raised for a schema validation error.
    """

    def __init__(self, msg, position=None):
        super(ValidationError, self).__init__(msg)
        self.position = position


class Container(object):

    """ Base class for all containers.
    """

    def populate_argparse(self, parser, name=None):
        pass

    def validate(self, value):
        raise NotImplementedError()


class ArgparseContainer(Container):

    """ A container which support argparse override.
    """

    def __init__(self, argparse_names=None, argparse_metavar=None,
                 argparse_help=None, argparse_names_invert=None,
                 argparse_help_invert=None):
        self._argparse_names = argparse_names
        self._argparse_value = None  # Where to store the overriding value
        self._argparse_metavar = argparse_metavar
        self._argparse_help = argparse_help
        self._argparse_names_invert = argparse_names_invert
        self._argparse_help_invert = argparse_help_invert


class Type(object):

    """ Base class for all types.
    """

    is_argparse_flag = False  # If True, this type will be defined as a
                              # boolean "flag" by argument parser.

    def validate(self, value):
        raise NotImplementedError()

    def cast(self, value):
        return value



""" Builtin types of confiture.schema
"""

import re
import sys
import numbers
import socket
import os.path

if sys.version_info[0] >= 3:
    import urllib.parse as urlparse
else:
    import urlparse

try:
    import ipaddr
except ImportError:
    IPADDR_ENABLED = False
else:
    IPADDR_ENABLED = True

from confiture.schema import Type, ValidationError


class Number(Type):

    """ A type representing a number (a float or an integer).
    """

    def validate(self, value):
        if not isinstance(value, numbers.Number):
            raise ValidationError('%r is not a number' % value)
        else:
            return value

    def cast(self, value):
        return float(value)


class Integer(Number):

    """ A type representing an integer in the configuration.

    :param min: define the minimum acceptable value value of the integer
    :param max: define the maximum acceptable value value of the integer

    Example in configuration::

        my_integer = 42
        my_integer = 42.0  # Will also match this type
    """

    def __init__(self, min=None, max=None):
        super(Integer, self).__init__()
        self._min = min
        self._max = max

    def validate(self, value):
        value = super(Integer, self).validate(value)
        if int(value) == value:
            value = int(value)
            if self._min is not None and value < self._min:
                raise ValidationError('%r is lower than the minimum (%d)'
                                      % (value, self._min))
            elif self._max is not None and value > self._max:
                raise ValidationError('%r is greater than the maximum (%d)'
                                      % (value, self._max))
            return value
        else:
            raise ValidationError('%r is not an integer value' % value)

    def cast(self, value):
        return int(value)


class Float(Number):

    """ A type representing a float in the configuration.

    Example in configuration::

        my_float = 42.2
        my_float = 42  # All values matched by the Integer type also match
                       # for the Float type
    """

    def __init__(self):
        super(Float, self).__init__()

    def validate(self, value):
        value = super(Float, self).validate(value)
        return float(value)


class Boolean(Type):

    """ A type representing a boolean value in the configuration.

    Example in configuration::

        my_boolean = yes
    """

    is_argparse_flag = True

    def __init__(self):
        super(Boolean, self).__init__()

    def validate(self, value):
        if value is not True and value is not False:
            raise ValidationError('%r is not a boolean value' % value)
        return value

    def cast(self, value):
        return bool(value)


class String(Type):

    """ A type representing a string in the configuration.

    Example in configuration::

        my_string = "hello, world!"

    """

    def __init__(self, encoding=None):
        super(String, self).__init__()
        self._encoding = encoding

    def validate(self, value):
        if self._encoding is not None:
            value = value.encode(self._encoding)
        return value

    def cast(self, value):
        return value


class _BaseRegex(String):

    """ Base class for regex types.
    """

    def __init__(self, regex, error='value doesn\'t match', **kwargs):
        super(_BaseRegex, self).__init__(**kwargs)
        self._regex = re.compile(regex)
        self._error = error

    def validate(self, value):
        value = super(_BaseRegex, self).validate(value)
        match = self._regex.match(value)
        if match is None:
            raise ValidationError(self._error)
        return match


class Regex(_BaseRegex):

    """ A string based type validated against a regex.
    """

    def validate(self, value):
        value = super(Regex, self).validate(value)
        return value.string


class NamedRegex(_BaseRegex):

    """ A string based type like Regex but returning named groups in dict.
    """

    def validate(self, value):
        value = super(NamedRegex, self).validate(value)
        return value.groupdict()


class RegexPattern(String):

    """ A re Python object.

    :param flags: python re compile `flag <http://docs.python.org/library/re.html#re.compile>`_

    Example in configuration::

        match = '/[a-z]+(-[a-z]+)?\.css'
    """

    def __init__(self, flags=0, **kwargs):
        super(RegexPattern, self).__init__(**kwargs)
        self.flags = flags

    def validate(self, value):
        value = super(RegexPattern, self).validate(value)
        # Try to compile regex object:
        try:
            value = re.compile(value, self.flags)
        except re.error:
            raise ValidationError('Bad format for regular expression')
        return value


class IPAddress(String):

    """ A string based type representing an ipv4 or ipv6 address.

    This type require the "ipaddr" package to work and will return an
    :class:`ipaddr.IPAddress` object.

    :param version: type or ip address to validate, can be 4 (ipv4 addresses
        only), 6 (ipv6 addresses only), or None (both).

    Example in configuration::

        interface = "127.0.0.1"
    """

    def __init__(self, version=None, **kwargs):
        if not IPADDR_ENABLED:
            raise RuntimeError('You must install the ipaddr package to use this type')
        super(IPAddress, self).__init__(**kwargs)
        self._version = version

    def validate(self, value):
        try:
            return ipaddr.IPAddress(value, version=self._version)
        except (ValueError, ipaddr.AddressValueError) as err:
            raise ValidationError(str(err))


class IPNetwork(IPAddress):

    """ A string based type representing an ipv4 or ipv6 network.

    This type require the "ipaddr" package to work and will return an
    :class:`ipaddr.IPNetwork` object.

    :param version: type or ip address to validate, can be 4 (ipv4 addresses
        only), 6 (ipv6 addresses only), or None (both).

    Example in configuration::

        allow = "10.0.0.0/8"
    """

    def validate(self, value):
        try:
            return ipaddr.IPNetwork(value, version=self._version)
        except ipaddr.AddressValueError:
            raise ValidationError('%r does not appear to be an IPv%s address'
                                  % (value, self._version))
        except ValueError as err:
            raise ValidationError(str(err))


class Url(String):

    """ A string based type representing an URL.

    This type return an urlparse.ParseResult object.

    Example in configuration::

        proxy = "http://proxy:3128"
    """

    def validate(self, value):
        try:
            return urlparse.urlparse(value)
        except ValueError as err:
            raise ValidationError(str(err))


class IPSocketAddress(String):

    """ A string based type representing an (ip address, port) couple.

    This type return an IPSocketAddress.Address object.

    Example in configuration::

        interface = "0.0.0.0:80"
    """

    def __init__(self, default_addr='127.0.0.1', default_port=None, version=None, **kwargs):
        if not IPADDR_ENABLED:
            raise RuntimeError('You must install the ipaddr package to use this type')
        self._default_addr = default_addr
        self._default_port = default_port
        self._version = version
        super(IPSocketAddress, self).__init__(**kwargs)

    def validate(self, value):
        raw_addr, _, raw_port = value.partition(':')
        if not raw_addr:
            raw_addr = self._default_addr
        if not raw_port:
            if self._default_port is None:
                raise ValidationError('You must specify a port')
            else:
                raw_port = self._default_port

        try:
            addr = ipaddr.IPAddress(raw_addr, version=self._version)
        except (ValueError, ipaddr.AddressValueError) as err:
            raise ValidationError(str(err))

        try:
            port = int(raw_port)
        except ValueError:
            raise ValidationError('%r is not a port (not an integer)' % raw_port)
        if not 1 <= port <= 65535:
            raise ValidationError('%r is not a port (not in 1 - 65535 range)' % port)

        return self.Address(addr, port)

    class Address(object):

        def __init__(self, addr, port):
            self.addr = addr
            self.port = port

        def __repr__(self):
            return 'IPSocketAddress.Address(%s:%s)' % (self.addr, self.port)

        def to_tuple(self):
            return (str(self.addr), self.port)

        def _create_socket(self, type):
            if self.addr.version == 4:
                family = socket.AF_INET
            else:
                family = socket.AF_INET6
            return socket.socket(family, type)

        def to_listening_socket(self, type):
            sock = self._create_socket(type)
            sock.listen(self.to_tuple())
            return sock

        def to_socket(self, type):
            sock = self._create_socket(type)
            sock.connect(self.to_tuple())
            return sock

        def to_listening_tcp_socket(self):
            return self.to_listening_socket(socket.SOCK_STREAM)

        def to_listening_udp_socket(self):
            return self.to_listening_socket(socket.SOCK_DGRAM)

        def to_tcp_socket(self):
            return self.to_socket(socket.SOCK_STREAM)

        def to_udp_socket(self):
            return self.to_socket(socket.SOCK_DGRAM)


class Eval(String):

    """ A string base type evaluating string as Python expression.

    Example in configuration::

        sum = 'sum(range(3, 10))'

    .. warning::
        This type can be dangerous since any Python expression can be typed
        by the user, like __import__("sys").exit(). Use it at your own risk.
    """

    def __init__(self, locals=None, globals=None, **kwargs):
        super(Eval, self).__init__(**kwargs)
        if locals is None:
            self._locals = {}
        else:
            self._locals = locals
        if globals is None:
            self._globals = {}
        else:
            self._globals = globals

    def validate(self, value):
        value = super(Eval, self).validate(value)
        try:
            return eval(value, self._globals, self._locals)
        except Exception as err:
            raise ValidationError('Bad expression: %s' % err)


class Path(String):

    """ A string representing a filesystem path.

    It will expand '~' to user home directory and return an absolute path if
    you provide a relative path (this is usefull if you change the working
    directory of a process after configuration parsing).
    """

    def validate(self, value):
        value = super(Path, self).validate(value)
        return os.path.abspath(os.path.expanduser(value))



""" Confiture lexer and parser.
"""

import sys
from glob import glob

import ply.lex as lex
import ply.yacc as yacc

from confiture.tree import ConfigSection, ConfigValue, Position


UNITS = {'k': 10 ** 3,
         'M': 10 ** 6,
         'G': 10 ** 9,
         'T': 10 ** 12,
         'P': 10 ** 15,
         'E': 10 ** 18,
         'Z': 10 ** 21,
         'Y': 10 ** 24,
         'Ki': 2 ** 10,
         'Mi': 2 ** 20,
         'Gi': 2 ** 30,
         'Ti': 2 ** 40,
         'Pi': 2 ** 50,
         'Ei': 2 ** 60,
         'Zi': 2 ** 70,
         'Yi': 2 ** 80}


class ParsingError(Exception):

    """ Error raised when a parsing error occurs.
    """

    def __init__(self, msg, position=None):
        super(ParsingError, self).__init__(msg)
        self.position = position


def default_external_opener(locator):
    """ The default locator used to open included external files.
    """
    parsed_externals = []
    for external in glob(locator):
        try:
            with open(external) as fexternal:
                external_data = fexternal.read()
        except IOError as err:
            raise ParsingError('Unable to open %s (%s)' % (external, err))
        parser = ConfitureParser(external_data, debug=False, write_tables=False,
                               errorlog=yacc.NullLogger(), input_name=external,
                               external_opener=default_external_opener)
        parsed_externals.append(parser.parse())
    return parsed_externals


#
# Lexer
#

class ConfitureLexer(object):

    """ Lexer for the DotConf format.

    :param \*\*kwargs: arguments to give to the ply lexer

    Usage example::

    >>> lexer = DotConfLexer()
    >>> lexer.input('test { key = yes }')
    >>> print lexer.next()
    """

    def __init__(self, encoding='utf-8', input_name='<unknown>', **kwargs):
        self._lexer = lex.lex(module=self, **kwargs)
        self._encoding = encoding
        self._input_name = input_name

    #
    # Tokens definition
    #

    reserved = {'yes': 'YES', 'no': 'NO', 'include': 'INCLUDE'}
    reserved.update(dict(((kw, 'UNIT') for kw in UNITS)))
    tokens = ['LBRACE', 'RBRACE', 'NAME', 'TEXT', 'NUMBER',
              'ASSIGN', 'LIST_SEP'] + list(set(reserved.values()))

    t_LBRACE = '{'
    t_RBRACE = '}'
    t_ASSIGN = '='
    t_LIST_SEP = ','
    t_ignore = ' \t'

    def t_NAME(self, token):
        r'[a-zA-Z_][a-zA-Z0-9_-]*'
        token.type = self.reserved.get(token.value, 'NAME')
        # Handle the boolean case:
        if token.type == 'YES':
            token.value = True
        elif token.type == 'NO':
            token.value = False
        elif token.type == 'UNIT':
            token.value = UNITS[token.value]
        return token

    def t_TEXT(self, token):
        r'(["]([\\]["]|[^"]|)*["]|[\']([\\][\']|[^\'])*[\'])'
        value = token.value[1:-1].replace('\\' + token.value[0], token.value[0])
        token.value = value
        # Count the lines in the string:
        token.lexer.lineno += value.count('\n')
        return token

    def t_NUMBER(self, token):
        r'[-+]?[0-9]+(\.[0-9]+)?'
        if token.value.isdigit():
            token.value = int(token.value)
        else:
            token.value = float(token.value)
        return token

    def t_EOL(self, token):
        r'[\n]+'
        token.lexer.lineno += len(token.value)

    def t_COMMENT(self, token):
        r'[#].*'

    def t_error(self, token):
        position = Position(self._input_name,
                            self._lexer.lineno,
                            self._lexer.lexpos)
        raise ParsingError('Illegal character %r' % token.value[0], position)

    #
    # Public methods
    #

    def column(self, lexpos):
        """ Find the column according to the lexpos.
        """
        # This code is taken from the python-ply documentation
        # see: http://www.dabeaz.com/ply/ply.html section 4.6
        last_cr = self._current_input.rfind('\n', 0, lexpos)
        if last_cr < 0:
            last_cr = 0
        column = (lexpos - last_cr)
        return column

    #
    # Bindings to the internal _lexer object
    #

    def input(self, input):
        if sys.version_info[0] >= 3:
            if isinstance(input, bytes):
                input = input.decode(self._encoding)
        else:
            if isinstance(input, str):
                input = input.decode(self._encoding)
        self._current_input = input
        return self._lexer.input(input)

    def __getattr__(self, name):
        attr = getattr(self._lexer, name)
        if attr is None:
            raise AttributeError("'%s' object has no attribute '%s'" % (self, name))
        else:
            return attr


#
# Parser
#

class ConfitureParser(object):

    """ Parser for the Confiture format.
    """

    tokens = ConfitureLexer.tokens

    def __init__(self, input, **kwargs):
        self._input = input
        self._input_name = kwargs.pop('input_name', '<unknown>')
        self._external_opener = kwargs.pop('external_opener',
                                           default_external_opener)
        self._lexer = kwargs.pop('lexer', ConfitureLexer(input_name=self._input_name))
        self._parser = yacc.yacc(module=self, **kwargs)
        self._old_line = 0

    def _check_line(self, current, lineno, pos, token):
        if self._old_line == current:
            pos = Position(self._input_name, lineno, pos)
            raise ParsingError('Syntax error near of "%s", '
                               'newline missing?' % token, pos)
        else:
            self._old_line = current

    #
    # Rules
    #

    start = 'top'

    def p_top(self, p):
        """top : section_content"""
        section = ConfigSection('__top__')
        for child in p[1]:
            if isinstance(child, ConfigSection):
                child.parent = section
            section.register(child)
        p[0] = section

    def p_assignation(self, p):
        """assignment : NAME ASSIGN value
                      | NAME ASSIGN list"""
        position = Position(self._input_name,
                            p.lineno(3),
                            p.lexer.column(p.lexpos(3)))
        value = ConfigValue(p[1], p[3], position=position)
        p[0] = value

    def p_value(self, p):
        """value : TEXT
                 | YES
                 | NO
                 | number"""
        p[0] = p[1]

    def p_number(self, p):
        """number : NUMBER
                  | NUMBER UNIT"""
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = p[1] * p[2]

    #
    # List management rules:
    #

    def p_list(self, p):
        """list : value LIST_SEP list_next
                | value LIST_SEP"""
        if len(p) == 3:
            p[0] = [p[1]]
        elif len(p) == 4:
            p[3].insert(0, p[1])
            p[0] = p[3]

    def p_list_next(self, p):
        """list_next : value
                     | value LIST_SEP
                     | value LIST_SEP list_next"""
        if len(p) in (2, 3):
            p[0] = [p[1]]
        elif len(p) == 4:
            p[3].insert(0, p[1])
            p[0] = p[3]

    #
    # Sections:
    #

    def p_section_content_empty(self, p):
        """section_content : empty"""
        p[0] = []

    def p_section_content_assignation(self, p):
        """section_content : section_content assignment
                           | section_content section"""
        self._check_line(p.lexer.lineno, p.lineno(2),
                         self._lexer.column(p.lexpos(2)), p[2].name)
        p[1].append(p[2])
        p[0] = p[1]

    def p_section_content_include(self, p):
        """section_content : section_content INCLUDE TEXT"""
        for external in self._external_opener(p[3]):
            p[1] += list(external.iterflatchildren())
        p[0] = p[1]

    def p_section(self, p):
        """section : NAME LBRACE section_content RBRACE
                   | NAME section_args LBRACE section_content RBRACE"""
        name = p[1]
        if len(p) == 5:
            args = None
            section_content = p[3]
        else:
            args = p[2]
            section_content = p[4]
        column = p.lexer.column(p.lexpos(1))
        position = Position(self._input_name, p.lineno(1), column)
        section = ConfigSection(name, args=args, position=position)
        for child in section_content:
            if isinstance(child, ConfigSection):
                child.parent = section
            section.register(child)
        p[0] = section

    def p_section_args(self, p):
        """section_args : value LIST_SEP section_args
                        | value"""
        if len(p) == 2:
            position = Position(self._input_name,
                                p.lineno(1),
                                p.lexer.column(p.lexpos(1)))
            p[0] = ConfigValue('<args>', [p[1]], position=position)
        else:
            p[3].value.insert(0, p[1])
            p[0] = p[3]

    def p_empty(self, p):
        """empty :"""
        pass

    def p_error(self, token):
        if token is None:
            raise ParsingError('Unexpected end of file')
        column = self._lexer.column(token.lexpos)
        pos = Position(self._input_name, token.lineno, column)
        raise ParsingError('Syntax error near of "%s"' % token.value, pos)

    #
    # Bindings to the internel _parser object
    #

    def parse(self):
        return self._parser.parse(self._input, self._lexer, tracking=True)

    def __getattr__(self, name):
        attr = getattr(self._parser, name)
        if attr is None:
            raise AttributeError("'%s' object has no attribute '%s'" % (self, name))
        else:
            return attr



""" Confiture's parser tests.
"""

import pytest

from confiture.parser import ConfitureLexer, ConfitureParser, ParsingError


def test_lexer():
    #          Test string            Expected tok     Expected tok value
    tokens = (('name',                'NAME',          'name'),
              ('"test"',              'TEXT',          'test'),
              ("'test'",              'TEXT',          'test'),
              (r"'te\'st'",           'TEXT',          "te'st"),
              ('42',                  'NUMBER',        42),
              ('42.1',                'NUMBER',        42.1),
              ('+42',                 'NUMBER',        42),
              ('+42.1',               'NUMBER',        42.1),
              ('-42',                 'NUMBER',        -42),
              ('-42.1',               'NUMBER',        -42.1),
              ('{',                   'LBRACE',        '{'),
              ('}',                   'RBRACE',        '}'),
              ('=',                   'ASSIGN',        '='))
    for test, expected_type, expected_value in tokens:
        yield check_token, test, expected_type, expected_value


def check_token(test, expected_type, expected_value):
    lexer = ConfitureLexer()
    lexer.input(test)
    token = lexer.next()
    assert token.type == expected_type
    assert token.value == expected_value

def test_parser_basic():
    test = '''
    daemon = yes  # This is a comment after an assignation
    # This is comment
    '''
    parser = ConfitureParser(test)
    output = parser.parse()
    assert output.get('daemon') is True

def test_parser_list():
    test = '''
    list1 = 1, 2, 3
    list2 = 1, 2, 3,
    list3 = 1,
    list4 = 1,
            2,
            3
    list5 = 1,
            2,
            3,
    '''
    parser = ConfitureParser(test)
    output = parser.parse()
    assert output.get('list1') == [1, 2, 3]
    assert output.get('list2') == [1, 2, 3]
    assert output.get('list3') == [1]
    assert output.get('list4') == [1, 2, 3]
    assert output.get('list5') == [1, 2, 3]

def test_parser_section():
    test = '''
    section1 {
        key = 'test'
    }
    section2 'arg' {}
    section3 'arg1', 'arg2' {}
    '''
    parser = ConfitureParser(test)
    output = parser.parse()
    assert tuple(output.subsections('section1'))[0].get('key') == 'test'
    assert tuple(output.subsections('section2'))[0].args == ['arg']
    assert tuple(output.subsections('section3'))[0].args == ['arg1', 'arg2']

def test_parser_empty():
    test = ''''''
    parser = ConfitureParser(test)
    output = parser.parse()

def test_parser_end_of_file():
    test = '''
    section {
    '''
    parser = ConfitureParser(test)
    with pytest.raises(ParsingError):
        parser.parse()



""" Builtin containers of confiture.schema
"""

import sys
try:
    import argparse
except ImportError:
    argparse = None

if sys.version_info[0] < 3:
    from itertools import izip as zip

from confiture.tree import ConfigSection, ConfigValue
from confiture.schema import Container, ArgparseContainer, ValidationError


required = object()
many = (1, None)
once = (1, 1)


class Value(ArgparseContainer):

    """ A value container used to store a scalar value of specified type.

    :param value_type: the type of the value stored by container
    :param default: the default value of the container
    """

    def __init__(self, value_type, default=required, **kwargs):
        super(Value, self).__init__(**kwargs)
        self._type = value_type
        self._default = default

    def populate_argparse(self, parser, name):
        value = self

        class Action(argparse.Action):

            def __init__(self, **kwargs):
                super(Action, self).__init__(**kwargs)
                self._const = kwargs.get('const', None)

            def __call__(self, parser, namespace, values, option_string=None):
                if self._const is not None:
                    value._argparse_value = ConfigValue(name, self._const)
                else:
                    value._argparse_value = ConfigValue(name, values[0])

        if self._argparse_names:
            if self._type.is_argparse_flag:
                nargs = 0
                const = True
            else:
                nargs = 1
                const = None
            nargs = 0 if self._type.is_argparse_flag else 1
            parser.add_argument(*self._argparse_names, action=Action,
                                type=self._type.cast, nargs=nargs,
                                metavar=self._argparse_metavar,
                                help=self._argparse_help, const=const)
            if self._type.is_argparse_flag and self._argparse_names_invert:
                parser.add_argument(*self._argparse_names_invert,
                                    action=Action, type=self._type.cast,
                                    nargs=0, help=self._argparse_help_invert,
                                    const=False)

    def validate(self, value):
        if self._argparse_value is not None:
            value = self._argparse_value
        if value is None:
            if self._default is required:
                raise ValidationError('this value is required')
            else:
                return ConfigValue(None, self._default)
        else:
            if isinstance(value.value, list):
                if len(value.value) == 1:
                    value.value = value.value[0]
                else:
                    raise ValidationError('%r is a list' % value.value,
                                          position=value.position)
            try:
                validated_value = self._type.validate(value.value)
            except ValidationError as err:
                raise ValidationError(str(err), position=value.position)
            return ConfigValue(value.name, validated_value, position=value.position)


class Choice(ArgparseContainer):

    """ A choice container used to store a choice of acceptable values.

    This container take a choices dict where each key is one of the acceptable
    values, and the according value, the value returned when the key is
    chosen.

    :param choices: the choices dict
    :param default: the default value of the container
    """

    def __init__(self, choices, default=required, **kwargs):
        super(Choice, self).__init__(**kwargs)
        self._choices = choices
        self._default = default

    def populate_argparse(self, parser, name):
        choice = self

        class Action(argparse.Action):

            def __init__(self, **kwargs):
                super(Action, self).__init__(**kwargs)
                self._const = kwargs.get('const', None)

            def __call__(self, parser, namespace, values, option_string=None):
                if self._const is not None:
                    choice._argparse_value = ConfigValue(name, self._const)
                else:
                    choice._argparse_value = ConfigValue(name, values)

        if self._argparse_names:
            parser.add_argument(*self._argparse_names, action=Action,
                                choices=list(self._choices.keys()),
                                metavar=self._argparse_metavar,
                                help=self._argparse_help)

    def validate(self, value):
        if self._argparse_value is not None:
            value = self._argparse_value
        if value is None:
            if self._default is required:
                raise ValidationError('this value is required')
            else:
                return ConfigValue(None, self._default)
        else:
            if isinstance(value.value, list):
                if len(value.value) == 1:
                    value.value = value.value[0]
                else:
                    raise ValidationError('%r is a list' % value.value,
                                          position=value.position)
            if value.value in self._choices:
                return ConfigValue(value.name, self._choices[value.value],
                                   position=value.position)
            else:
                choices = ', '.join(repr(x) for x in self._choices)
                raise ValidationError('bad choice (must be one of %s)' % choices)


class List(ArgparseContainer):

    """ A list container used to store a list of scalar value of specified type.

    :param values_type: type of values
    :param default: the default value of the container
    """

    def __init__(self, values_type, default=required, **kwargs):
        super(List, self).__init__(**kwargs)
        self._type = values_type
        self._default = default

    def populate_argparse(self, parser, name):
        value = self

        class Action(argparse.Action):

            def __call__(self, parser, namespace, values, option_string=None):
                value._argparse_value = ConfigValue(name, values)

        if self._argparse_names:
            nargs = '*'
            parser.add_argument(*self._argparse_names, action=Action,
                                type=self._type.cast, nargs=nargs,
                                metavar=self._argparse_metavar,
                                help=self._argparse_help)

    def validate(self, value):
        if self._argparse_value is not None:
            value = self._argparse_value
        if value is None:
            if self._default is required:
                raise ValidationError('this value is required')
            else:
                return ConfigValue(None, self._default)
        else:
            values = value.value
            if not isinstance(values, list):
                values = [values]
            validated_list = []
            for i, item in enumerate(values):
                try:
                    item = self._type.validate(item)
                except ValidationError as err:
                    raise ValidationError('item #%d, %s' % (i, err),
                                          position=value.position)
                else:
                    validated_list.append(item)
            return ConfigValue(value.name, validated_list, position=value.position)


class Array(List):

    """ An array container used to store a fixed size list of scalar values of
        the specified type.

    :param size: size of the array
    :param \*\*kwargs: same arguments as List
    """

    def __init__(self, size, *args, **kwargs):
        super(Array, self).__init__(*args, **kwargs)
        self._size = size

    def populate_argparse(self, parser, name):
        value = self

        class Action(argparse.Action):

            def __call__(self, parser, namespace, values, option_string=None):
                value._argparse_value = ConfigValue(name, values)

        if self._argparse_names:
            parser.add_argument(*self._argparse_names, action=Action,
                                type=self._type.cast, nargs=self._size,
                                metavar=self._argparse_metavar,
                                help=self._argparse_help)

    def validate(self, value):
        value = super(Array, self).validate(value)
        if len(value.value) != self._size:
            raise ValidationError('bad array size (should be %d, found %d items)'
                                  % (self._size, len(value.value)))
        return value


class TypedArray(ArgparseContainer):

    """ An array container used to store a fixed size list of scalar values
        with specified type for each of them.

    :param values_types: types of each item in a list
    :param default: the default value of the container
    """

    def __init__(self, values_types, default=required, **kwargs):
        super(TypedArray, self).__init__(**kwargs)
        self._types = values_types
        self._default = default

    def populate_argparse(self, parser, name):
        value = self

        class Action(argparse.Action):

            def __call__(self, parser, namespace, values, option_string=None):
                value._argparse_value = ConfigValue(name, values)

        if self._argparse_names:
            nargs = '*'
            parser.add_argument(*self._argparse_names, action=Action,
                                type=str, nargs=nargs,
                                metavar=self._argparse_metavar,
                                help=self._argparse_help)

    def validate(self, value):
        if self._argparse_value is not None:
            value = self._argparse_value
        if value is None:
            if self._default is required:
                raise ValidationError('this value is required')
            else:
                return ConfigValue(None, self._default)
        else:
            values = value.value
            if not isinstance(values, list):
                values = [values]
            validated_list = []
            if len(values) != len(self._types):
                raise ValidationError('bad array size (should be %d, found %d '
                                      'items)' % (len(self._types), len(values)))

            for i, (item, item_type) in enumerate(zip(values, self._types)):
                try:
                    item = item_type.validate(item)
                except ValidationError as err:
                    raise ValidationError('item #%d, %s' % (i, err),
                                          position=value.position)
                else:
                    validated_list.append(item)
            return ConfigValue(value.name, validated_list, position=value.position)


class Section(Container):

    """ A section container used to store a mapping between name and other
        containers.

        :param \*\*kwargs: parameters used to override meta
    """

    # Defaults meta:
    _meta = {'args': None,
             'unique': False,
             'repeat': once,
             'allow_unknown': False}

    def __init__(self, **kwargs):
        self.meta = {}
        self.keys = {}
        for cls in reversed(self.__class__.mro()):
            # Update meta from class:
            if hasattr(cls, '_meta'):
                self.meta.update(cls._meta)
            # Update fields from class:
            for key, value in cls.__dict__.items():
                if isinstance(value, Container):
                    self.keys[key] = value

    def add(self, name, container):
        """ Add a new key imperatively.

        :param name: the key name to add
        :param container: the container to add
        """

        if name not in self.keys:
            self.keys[name] = container
        else:
            raise KeyError('key already exists')

    def populate_argparse(self, parser, name=None):
        """ Populate an argparse parser.
        """

        for name, container in self.keys.items():
            container.populate_argparse(parser, name=name)

    def validate(self, section):
        if not isinstance(section, ConfigSection):
            raise ValidationError('Not a section')

        # Rebuild the section using schema:
        validated_section = ConfigSection(section.name, section.parent,
                                          position=section.position)
        # Validate the section's argument:
        if self.meta['args'] is None and section.args is not None:
            raise ValidationError('section %s, this section does not take '
                                  'any argument' % section.name,
                                  position=section.position)
        elif self.meta['args'] is not None:
            try:
                validated_args = self.meta['args'].validate(section.args_raw)
            except ValidationError as err:
                msg = 'section %s, arguments, %s' % (section.name, err)
                raise ValidationError(msg, position=err.position)
            else:
                validated_section.args = validated_args
        # Validate the section's children:
        for name, container in self.keys.items():
            if isinstance(container, Section):
                # Validate subsections of this section:
                subsections = list(section.subsections(name))
                # Check for repeat option:
                rmin, rmax = container.meta['repeat']
                if rmax is not None and rmin > rmax:
                    raise ValidationError('section %s, rmin > rmax' % name)
                if len(subsections) < rmin:
                    raise ValidationError('section %s, section must be defined'
                                          ' at least %d times' % (name, rmin))
                if rmax is not None and len(subsections) > rmax:
                    raise ValidationError('section %s, section must be defined'
                                          ' at max %d times' % (name, rmax))
                # Do the children validation:
                args = set()  # Store the already seen args
                for subsection in subsections:
                    # Check for unique option:
                    if container.meta['unique']:
                        args_value = None if subsection.args is None else tuple(subsection.args)
                        if args_value in args:
                            msg = 'section %s, section must be unique' % name
                            raise ValidationError(msg, position=subsection.position)
                        else:
                            args.add(args_value)
                    # Container validation:
                    validated_subsection = container.validate(subsection)
                    validated_section.register(validated_subsection, name=name)
            elif isinstance(container, Container):
                # Validate all other types of containers:
                try:
                    validated_value = container.validate(section.get(name, raw=False))
                except ValidationError as err:
                    raise ValidationError('section %s, key %s, %s' % (section.name, name, err),
                                          position=err.position)
                else:
                    validated_section.register(validated_value, name=name)
        # Handle the allow_unknown meta option:
        for name, child in section.iteritems(expand_sections=True):
            if name not in self.keys:
                if self.meta['allow_unknown']:
                    validated_section.register(child, name=name)
                else:
                    msg = 'section %s, unknown key %s' % (section.name, name)
                    raise ValidationError(msg, position=child.position)
        return validated_section



import sys
from pprint import pprint

from confiture import Confiture
from confiture.schema import ValidationError
from confiture.parser import ParsingError
from confiture.schema.containers import Section, Value, many
from confiture.schema.types import Boolean, Integer, String

#
# Our Confiture schemas:
#

class GenericPluginSchema(Section):

    # This value is common to all plugins:
    common_value = Value(Integer())

    # Enable the allow_unknown meta in order to keep unknown values
    # on the first validation:
    _meta = {'allow_unknown': True,
             'repeat': many,
             'args': Value(String())}


class MainConfigurationSchema(Section):

    global_value = Value(Integer())
    plugin = GenericPluginSchema()


#
# Our plugins:
#

class BasePlugin(object):

    name = None
    schema = None

    def __init__(self, config):
        self.config = config

    def print_config(self):
        print '%s: ' % self.__class__.__name__
        pprint(self.config.to_dict())


class BeautifulSchema(GenericPluginSchema):

    beautiful_value = Value(String())

    # The allow_unknown meta is disable to forbid bad config:
    _meta = {'allow_unknown': False}


class BeautifulPlugin(BasePlugin):

    name = 'beautiful'
    schema = BeautifulSchema()


class UglySchema(GenericPluginSchema):

    ugly_value = Value(Boolean())
    _meta = {'allow_unknown': False}


class UglyPlugin(BasePlugin):

    name = 'ugly'
    schema = UglySchema()



#
# Our test config
#

TEST_CONFIG = '''
global_value = 42

plugin 'beautiful' {
    common_value = 123
    beautiful_value = "I'm so beautiful"
}

plugin 'ugly' {
    common_value = 456
    ugly_value = no
}
'''

#
# This is where the magic happen:
#

if __name__ == '__main__':
    my_plugins = (BeautifulPlugin, UglyPlugin)
    enabled_plugins = []

    # Parse the global configuration:
    config = Confiture(TEST_CONFIG, schema=MainConfigurationSchema())
    try:
        pconfig = config.parse()
    except (ValidationError, ParsingError) as err:
        if err.position is not None:
            print str(err.position)
        print err
        sys.exit(1)
    else:
        print 'Main configuration:'
        pprint(pconfig.to_dict())

    # Enable each used plugins:
    for plugin_conf in pconfig.subsections('plugin'):
        # Search the plugin:
        for plugin in my_plugins:
            if plugin.name == plugin_conf.args:
                break
        else:
            print 'Unknown plugin %r, exiting.' % plugin_conf.args
            sys.exit(1)
        # Check plugin configuration:
        try:
            validated_conf = plugin.schema.validate(plugin_conf)
        except ValidationError as err:
            print 'Bad plugin configuration:'
            if err.position is not None:
                print str(err.position)
            print err
            sys.exit(1)
        else:
            # Instanciate the plugin object:
            enabled_plugins.append(plugin(validated_conf))

    # Print each enabled plugin config:
    for plugin in enabled_plugins:
        print '\n' + '~' * 80 + '\n'
        plugin.print_config()




""" Confiture is an advanced configuration parser for Python.
"""


from confiture.parser import ConfitureParser, yacc


class Confiture(object):

    def __init__(self, config, schema=None, input_name='<unknown>'):
        self._config = config
        self._schema = schema
        self._input_name = input_name

    @classmethod
    def from_filename(cls, filename, **kwargs):
        fconf = open(filename, 'Ur')
        kwargs['input_name'] = filename
        return cls(fconf.read(), **kwargs)

    def _parse(self):
        parser = ConfitureParser(self._config, debug=False, write_tables=False,
                               errorlog=yacc.NullLogger(), input_name=self._input_name)

        return parser.parse()

    def parse(self):
        config = self._parse()
        if self._schema is not None:
            config = self._schema.validate(config)
        return config



from setuptools import setup, find_packages
from io import open
import os

version = '2.1'
ldesc = open(os.path.join(os.path.dirname(__file__), 'README.rst'), encoding='utf8').read()

setup(name='confiture',
      version=version,
      description='Advanced configuration parser for Python',
      long_description=ldesc,
      classifiers=['Programming Language :: Python',
                   'Programming Language :: Python :: 3',
                   'Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Topic :: Software Development',
                   'Topic :: Software Development :: Libraries',
                   'Topic :: Text Processing'],
      keywords='configuration parser confiture conf dotconf',
      author='Antoine Millet',
      author_email='antoine@inaps.org',
      url='https://idevelop.org/p/confiture/',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      install_requires=['ply'])



""" Classes used to represent the configuration tree.
"""


from itertools import chain
from collections import defaultdict


class MultipleSectionsWithThisNameError(Exception):
    """ Exception raised if only one section is expected, but multiple returned.
    """


class Position(object):

    """ Position of a statement in a file.
    """

    def __init__(self, file_, lineno, pos):
        self.file = file_
        self.lineno = lineno
        self.pos = pos

    def __repr__(self):
        return '<Position file=%s lineno=%s pos=%s>' % (self.file, self.lineno, self.pos)

    def __str__(self):
        return 'in %s, line %d, position %d' % (self.file, self.lineno, self.pos)


class ConfigValue(object):

    """ Represent a value in the configuration.
    """

    def __init__(self, name, value, position=Position('?', 0, 0)):
        self._name = name
        self._value = value
        self._position = position

    def __repr__(self):
        return '<ConfigValue %r (%s)>' % (self._value, self._position)

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def position(self):
        return self._position


class ConfigSection(object):

    """ Represent a (sub-)section of the configuration.

    :param name: the name of the section (or __top__ for top section)
    :param parent: the parent section (or None for top section)
    :param position: the position of the section in configuration
    """

    def __init__(self, name, parent=None, args=None, position=Position('?', 0, 0)):
        self._name = name
        self._parent = parent
        self._args = args
        self._position = position
        self._subsections = defaultdict(lambda: [])
        self._values = {}

    def __repr__(self):
        return "<Section '%s'>" % self.name

    def __contains__(self, name):
        return name in self._values or name in self._subsections

    #
    # Public API -- Tree construction methods
    #   These methods are intended as internal usage for the configuration
    #   tree construction by the parser or the schema system.
    #

    def register(self, child, name=None):
        """ Register a child on this section.

        :param child: ConfigValue or ConfigSection object to register
        """
        if name is None:
            name = child.name
        if isinstance(child, ConfigValue):
            if name in self:
                raise KeyError('A child with this name already exists')
            self._values[name] = child
        elif isinstance(child, ConfigSection):
            if name in self._values:
                raise KeyError('A child with this name already exists')
            self._subsections[name].append(child)
        else:
            raise TypeError('Child must be a ConfigValue or ConfigSection object')

    def iterchildren(self):
        """ Iterate over all children of this section.
        """

        return chain(iter(self._values.values()), iter(self._subsections.values()))

    def iterflatchildren(self):
        """ Iterate over all children of this section, not grouping sections.

        ..note::

            This is used for the include feature of parser.
        """
        return chain(iter(self._values.values()), *iter(self._subsections.values()))

    def iteritems(self, expand_sections=False):
        """ Like :meth:``iterchildren`` but return a couple (key, child).

        If expand_sections is True, the method will return a couple for each
        occurrence of a section instead of returning a couple with a list
        of section.
        """

        if expand_sections:
            return chain(iter(self._values.items()),
                         # Expand {'x': [1, 2]}.items() into ((x, 1), (x, 2)):
                         ((k, v) for k, l in self._subsections.items() for v in l))
        else:
            return chain(iter(self._values.items()), iter(self._subsections.items()))

    #
    # Public API -- User methods
    #   These methods are intended for configuration tree traversal by user.
    #

    @property
    def name(self):
        return self._name

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    @property
    def args(self):
        if self._args is not None:
            return self._args.value
        else:
            return None

    @args.setter
    def args(self, value):
        self._args = value

    @property
    def args_raw(self):
        return self._args

    @property
    def position(self):
        return self._position

    def subsections(self, name):
        """ Iterate over sub-sections with the specified name.
        """
        return iter(self._subsections[name])

    def subsection(self, name, default=None):
        """ Get sub-section with the specified name.

        Work like subsections method, but can only be applied if there is
        exactly one subsection with the specified name present. Otherwise,
        a MultipleSectionsWithThisNameError exception is raised. This function
        is useful if you do not need the list support for sections.
        """
        if len(self._subsections.get(name, [])) > 1:
            msg = '%s.subsection can\'t return multiple sections' % self.__class__.__name__
            raise MultipleSectionsWithThisNameError(msg)
        elif name not in self._subsections or len(self._subsections[name]) == 0:
            return default
        else:
            return self._subsections[name][0]

    def get(self, name, default=None, raw=True):
        """ Get the value with the specified name or return default.

        :param default: the default value to return if the name is not found
        :param raw: get the ConfigValue object instead of raw value

        .. note::
           If the default value is returned, the raw option is implicitly
           enabled.
        """
        value = self._values.get(name, default)
        if value is not default and raw:
            value = value.value
        return value

    def to_dict(self):
        """ Represent the section (and subsections) as a dict.
        """
        output = {}
        for name, subsections in self._subsections.items():
            output[name] = []
            for subsection in subsections:
                output[name].append(subsection.to_dict())
        for name, value in self._values.items():
            output[name] = value.value
        return output
